#include "GraspProviderExample.h"

#include <SimoxUtility/color/cmaps.h>

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>
#include <ArmarXCore/core/time/CycleUtil.h>

#include <RobotAPI/libraries/core/Pose.h>

#include <RobotAPI/libraries/armem/server/MemoryRemoteGui.h>
#include <RobotAPI/libraries/armem/client/query/Builder.h>
#include <RobotAPI/libraries/armem/client/query/query_fns.h>
#include <RobotAPI/libraries/armem/core/ice_conversions.h>

#include <RobotAPI/libraries/GraspingUtility/aron/GraspCandidate.aron.generated.h>


namespace armarx
{
    GraspProviderExamplePropertyDefinitions::GraspProviderExamplePropertyDefinitions(std::string prefix) :
        armarx::ComponentPropertyDefinitions(prefix)
    {
    }

    armarx::PropertyDefinitionsPtr GraspProviderExample::createPropertyDefinitions()
    {
        ARMARX_IMPORTANT << "Prperty defs";
        armarx::PropertyDefinitionsPtr defs = new GraspProviderExamplePropertyDefinitions(getConfigIdentifier());

        defs->topic(debugObserver);

        defs->optional(memoryName, "mem.MemoryName", "Name of the memory to use.");

        return defs;

    }

    std::string GraspProviderExample::getDefaultName() const
    {
        return "GraspProviderExample";
    }

    GraspProviderExample::GraspProviderExample() : writer(memoryNameSystem()), reader(memoryNameSystem())
    {

    }


    void GraspProviderExample::onInitComponent()
    {
        ARMARX_IMPORTANT << "Init";
    }


    void GraspProviderExample::onConnectComponent()
    {
        writer.connect();
        reader.connect();
        task = new RunningTask<GraspProviderExample>(this, &GraspProviderExample::run);
        task->start();
    }


    void GraspProviderExample::onDisconnectComponent()
    {
        task->stop();
    }


    void GraspProviderExample::onExitComponent()
    {
    }


    void GraspProviderExample::run()
    {
        ARMARX_IMPORTANT << "Running example.";

        CycleUtil c(1000);
        int i = 0;


        while (!task->isStopped() && i++ < 100)
        {
            // initialize all necessary fields of a grasp candidate and use writer to commit it to memory

            armarx::grasping::GraspCandidate candidate = armarx::grasping::GraspCandidate();
            candidate.groupNr = i; //non-necessary field, but used to commit different candidates
            candidate.approachVector = Vector3BasePtr(toIce(Eigen::Vector3f()));
            candidate.graspPose = PoseBasePtr(toIce(Eigen::Matrix4f()));
            candidate.providerName = "Example";
            candidate.robotPose = PoseBasePtr(toIce(Eigen::Matrix4f()));
            candidate.tcpPoseInHandRoot = PoseBasePtr(toIce(Eigen::Matrix4f()));
            // source Info is also not necessary, but reference object name is used as entity name
            // "UnknownObject" if none is provided
            candidate.sourceInfo = new grasping::GraspCandidateSourceInfo();
            candidate.sourceInfo->referenceObjectName = "Box";
            candidate.sourceInfo->bbox = new grasping::BoundingBox();
            candidate.sourceInfo->bbox->center = Vector3BasePtr(toIce(Eigen::Vector3f()));
            candidate.sourceInfo->bbox->ha1 = Vector3BasePtr(toIce(Eigen::Vector3f()));
            candidate.sourceInfo->bbox->ha2 = Vector3BasePtr(toIce(Eigen::Vector3f()));
            candidate.sourceInfo->bbox->ha3 = Vector3BasePtr(toIce(Eigen::Vector3f()));
            candidate.sourceInfo->referenceObjectPose = PoseBasePtr(toIce(Eigen::Matrix4f()));

            writer.commitGraspCandidate(candidate, armem::Time::Now(), "provider1");
            // initialize all necessary fields of a bimanual grasp candidate and use writer to commit it to memory
            armarx::grasping::BimanualGraspCandidate bimanualCandidate = armarx::grasping::BimanualGraspCandidate();
            bimanualCandidate.groupNr = i; //non-necessary field, but used to commit different candidates
            bimanualCandidate.approachVectorLeft = Vector3BasePtr(toIce(Eigen::Vector3f()));
            bimanualCandidate.approachVectorRight = Vector3BasePtr(toIce(Eigen::Vector3f()));
            bimanualCandidate.graspPoseLeft = PoseBasePtr(toIce(Eigen::Matrix4f()));
            bimanualCandidate.graspPoseRight = PoseBasePtr(toIce(Eigen::Matrix4f()));
            bimanualCandidate.providerName = "BimanualExample";
            bimanualCandidate.robotPose = PoseBasePtr(toIce(Eigen::Matrix4f()));
            bimanualCandidate.tcpPoseInHandRootRight = PoseBasePtr(toIce(Eigen::Matrix4f()));
            bimanualCandidate.tcpPoseInHandRootLeft = PoseBasePtr(toIce(Eigen::Matrix4f()));
            bimanualCandidate.inwardsVectorLeft = Vector3BasePtr(toIce(Eigen::Vector3f()));
            bimanualCandidate.inwardsVectorRight = Vector3BasePtr(toIce(Eigen::Vector3f()));

            writer.commitBimanualGraspCandidate(bimanualCandidate, armem::Time::Now(), "provider2");

            //test for writing Seqs, candidates from the same object appear as instances of the same snapshot
            grasping::GraspCandidateSeq candidatesToWrite;
            candidatesToWrite.push_back(new grasping::GraspCandidate(candidate));
            candidate.side = "Left";

            candidatesToWrite.push_back(new grasping::GraspCandidate(candidate));

            writer.commitGraspCandidateSeq(candidatesToWrite, armem::Time::Now(), "provider1");

            // test reader and debug by logging the group number of the candidate

            std::map<std::string, grasping::GraspCandidatePtr> candidates;

            try
            {
                candidates = reader.queryLatestGraspCandidates();
            }
            catch (armem::error::QueryFailed &e)
            {
                ARMARX_ERROR << e.makeMsg(memoryName);
            }


            for (auto [id, ca] : candidates)
            {
                ARMARX_INFO << "candidate with ID " << id << " has group number " <<  ca->groupNr ;
            }

            std::map<std::string, grasping::BimanualGraspCandidatePtr> bimanualCandidates;

            try
            {
                bimanualCandidates = reader.queryLatestBimanualGraspCandidates();
            }
            catch (armem::error::QueryFailed &e)
            {
                ARMARX_ERROR << e.makeMsg(memoryName);
            }

            for (auto [id, ca] : bimanualCandidates)
            {
                ARMARX_INFO << "bimanual candidate with ID " << id << " has group number " <<  ca->groupNr ;
            }

            c.waitForCycleDuration();
        }
    }

}
