#include "Segment.h"

// STL
#include <iterator>

#include <ArmarXCore/core/logging/Logging.h>

#include <RobotAPI/libraries/core/FramedPose.h>
#include <RobotAPI/libraries/aron/common/aron_conversions.h>

#include <RobotAPI/libraries/armem/core/MemoryID.h>
#include <RobotAPI/libraries/armem/core/Time.h>
#include <RobotAPI/libraries/armem/core/aron_conversions.h>
#include <RobotAPI/libraries/armem/server/MemoryToIceAdapter.h>

#include <RobotAPI/libraries/armem_robot/aron/Robot.aron.generated.h>
#include <RobotAPI/libraries/armem_robot/robot_conversions.h>

#include <RobotAPI/libraries/armem_robot_state/aron_conversions.h>
#include <RobotAPI/libraries/armem_robot_state/types.h>
#include <RobotAPI/libraries/armem_robot_state/common/localization/TransformHelper.h>
#include <RobotAPI/libraries/armem_robot_state/common/localization/types.h>


namespace armarx::armem::server::robot_state::localization
{

    Segment::Segment(armem::server::MemoryToIceAdapter& memoryToIceAdapter) :
        Base(memoryToIceAdapter, "Localization", arondto::Transform::ToAronType(), 1024)
    {
    }


    Segment::~Segment()
    {
    }


    void Segment::onConnect()
    {
    }


    RobotFramePoseMap
    Segment::getRobotFramePosesLocking(const armem::Time& timestamp) const
    {
        return this->doLocked([this, &timestamp]()
        {
            return getRobotFramePoses(timestamp);
        });
    }


    RobotFramePoseMap
    Segment::getRobotFramePoses(const armem::Time& timestamp) const
    {
        using common::robot_state::localization::TransformHelper;
        using common::robot_state::localization::TransformQuery;

        RobotFramePoseMap frames;
        for (const std::string& robotName : segmentPtr->getProviderSegmentNames())
        {
            TransformQuery query
            {
                .header = {
                    .parentFrame = armarx::GlobalFrame,
                    .frame       = "root", // TODO, FIXME
                    .agent       = robotName,
                    .timestamp   = timestamp
                }
            };

            const auto result = TransformHelper::lookupTransformChain(*segmentPtr, query);
            if (not result)
            {
                // TODO
                continue;
            }

            frames.emplace(robotName, result.transforms);
        }

        ARMARX_INFO << deactivateSpam(60)
                    << "Number of known robot pose chains: " << frames.size();

        return frames;
    }


    RobotPoseMap
    Segment::getRobotGlobalPosesLocking(const armem::Time& timestamp) const
    {
        return this->doLocked([this, &timestamp]()
        {
            return getRobotGlobalPoses(timestamp);
        });
    }


    RobotPoseMap
    Segment::getRobotGlobalPoses(const armem::Time& timestamp) const
    {
        using common::robot_state::localization::TransformHelper;
        using common::robot_state::localization::TransformQuery;

        RobotPoseMap robotGlobalPoses;
        for (const std::string& robotName : segmentPtr->getProviderSegmentNames())
        {
            TransformQuery query
            {
                .header =
                {
                    .parentFrame = GlobalFrame,
                    .frame       = "root", // TODO, FIXME
                    .agent       = robotName,
                    .timestamp   = timestamp
                }
            };

            if (const auto result = TransformHelper::lookupTransform(*segmentPtr, query))
            {
                robotGlobalPoses.emplace(robotName, result.transform.transform);
            }
            else
            {
                // TODO
            }
        }

        ARMARX_INFO << deactivateSpam(60)
                    << "Number of known robot poses: " << robotGlobalPoses.size();

        return robotGlobalPoses;
    }


    bool Segment::commitTransform(const armarx::armem::robot_state::Transform& transform)
    {
        Commit commit;
        commit.add(makeUpdate(transform));

        const armem::CommitResult result = iceMemory.commit(commit);
        return result.allSuccess();
    }


    bool Segment::commitTransformLocking(const armarx::armem::robot_state::Transform& transform)
    {
        Commit commit;
        commit.add(makeUpdate(transform));

        const armem::CommitResult result = iceMemory.commitLocking(commit);
        return result.allSuccess();
    }


    EntityUpdate Segment::makeUpdate(const armarx::armem::robot_state::Transform& transform) const
    {
        const armem::Time& timestamp = transform.header.timestamp;
        const MemoryID providerID = segmentPtr->id().withProviderSegmentName(transform.header.agent);

        EntityUpdate update;
        update.entityID = providerID.withEntityName(transform.header.parentFrame + "," + transform.header.frame);
        update.timeArrived = update.timeCreated = update.timeSent = timestamp;

        arondto::Transform aronTransform;
        toAron(aronTransform, transform);
        update.instancesData = {aronTransform.toAron()};

        return update;
    }

} // namespace armarx::armem::server::robot_state::localization
