/**
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Reister ( fabian dot reister at kit dot edu )
 * @date       2022
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "PeriodicSkill.h"

#include "ArmarXCore/core/exceptions/LocalException.h"
#include "ArmarXCore/core/logging/Logging.h"
#include "ArmarXCore/core/time/Frequency.h"
#include <ArmarXCore/core/time/Metronome.h>

#include "RobotAPI/libraries/skills/provider/Skill.h"

namespace armarx::skills
{
    PeriodicSkill::PeriodicSkill(const SkillDescription& skillDescription,
                                 const armarx::Frequency& frequency) :
        Skill(skillDescription), frequency(frequency)
    {
    }

    Skill::Status
    PeriodicSkill::execute(const aron::data::DictPtr& params, const CallbackT& callback)
    {
        if(not initialize(params))
        {
            onFailed();
            return Skill::Status::Failed;
        }

        core::time::Metronome metronome(frequency);

        while (not stopped and not timeoutReached)
        {
            const auto status = executeOnce(params);
            switch (status)
            {
                case Status::Running:
                    // nothing to do here
                    break;
                case Status::Succeeded:
                    onSucceeded();
                    return Skill::Status::Succeeded;
                case Status::Failed:
                    onFailed();
                    return Skill::Status::Failed;
            }

            const auto sleepDuration = metronome.waitForNextTick();
            if (not sleepDuration.isPositive())
            {
                ARMARX_INFO << deactivateSpam() << "PeriodicSkill: execution took too long ("
                            << -sleepDuration << " vs " << frequency.toCycleDuration() << ")";
            }
        }

        if (stopped)
        {
            onStopped();
            return Skill::Status::Stopped;
        }

        if (timeoutReached)
        {
            onTimeoutReached();
            return Skill::Status::TimeoutReached;
        }

        throw armarx::LocalException("should not happen.");
    }


} // namespace armarx::skills
