#include "SkillProviderComponentPlugin.h"

#include <ArmarXCore/core/Component.h>

#include <RobotAPI/libraries/aron/core/data/variant/primitive/All.h>

namespace armarx::plugins
{
    void SkillProviderComponentPlugin::preOnInitComponent()
    {

    }

    void SkillProviderComponentPlugin::preOnConnectComponent()
    {
        auto& p = parent<SkillProviderComponentPluginUser>();
        p.getProxy(myPrx, -1);
    }

    void SkillProviderComponentPlugin::postOnConnectComponent()
    {
        auto& p = parent<SkillProviderComponentPluginUser>();
        std::string providerName = p.getName();

        // update skill ownership
        for (auto& [skillName, impl] : p.skillImplementations)
        {
            impl.skill->ownerProvider = myPrx;
            impl.skill->ownerManager = ownedBySkillManager;

            impl.statusUpdate.providerName = p.getName();
            impl.statusUpdate.skillName = skillName;
        }

        // register to manager
        skills::manager::dto::ProviderInfo i;
        i.provider = myPrx;
        i.providerName = providerName;
        i.providedSkills = p.getSkills();
        ownedBySkillManager->addProvider(i);

        p.connected = true;
    }

    void SkillProviderComponentPlugin::preOnDisconnectComponent()
    {
        auto& p = parent<SkillProviderComponentPluginUser>();
        std::string providerName = p.getName();

        ownedBySkillManager->removeProvider(providerName);
    }

    void SkillProviderComponentPlugin::postCreatePropertyDefinitions(PropertyDefinitionsPtr& properties)
    {
        std::string prefix = "skill.";
        properties->component(ownedBySkillManager, "SkillMemory", prefix + "SkillManager", "The name of the SkillManager (or SkillMemory) proxy this provider belongs to.");
    }
}


namespace armarx
{
    SkillProviderComponentPluginUser::SkillProviderComponentPluginUser()
    {
        addPlugin(plugin);
    }

    void SkillProviderComponentPluginUser::addSkill(std::unique_ptr<skills::Skill>&& skill)
    {
        if (!skill)
        {
            return;
        }

        if (connected)
        {
            ARMARX_WARNING << "The SkillProvider already registered to a manager. The skill '" + skill->description.skillName + "' therefore cannot be added anymore.";
            return;
        }

        std::unique_lock l(skillsMutex);

        std::string skillName = skill->description.skillName;
        if (skillImplementations.find(skillName) != skillImplementations.end())
        {
            ARMARX_WARNING << "Try to add a skill '" + skillName + "' which already exists in list. Ignoring this skill.";
            return;
        }

        ARMARX_INFO << "Adding skill " << skillName;
        skillImplementations.emplace(skillName, std::move(skill));
    }

    void SkillProviderComponentPluginUser::addSkill(const skills::helper::LambdaSkill::FunT& f, const skills::SkillDescription& desc)
    {
        auto lambda = std::make_unique<skills::helper::LambdaSkill>(f, desc);
        addSkill(std::move(lambda));
    }

    skills::provider::dto::SkillDescriptionMap SkillProviderComponentPluginUser::getSkills(const Ice::Current &current)
    {
        std::shared_lock l(skillsMutex);
        skills::provider::dto::SkillDescriptionMap skillDesciptions;
        for (const auto& [key, skillWrapper] : skillImplementations)
        {
            skillDesciptions.insert({key, skillWrapper.skill->description.toIce()});
        }
        return skillDesciptions;
    }

    skills::provider::dto::SkillStatusUpdate SkillProviderComponentPluginUser::getSkillExecutionStatus(const std::string& skill, const Ice::Current &current)
    {
        std::shared_lock l(skillsMutex);
        auto& skillWrapper = skillImplementations.at(skill);

        std::shared_lock l2(skillWrapper.skillStatusMutex);
        return skillWrapper.statusUpdate.toIce();
    }

    // Please not that this method waits until the skill can be scheduled!
    void SkillProviderComponentPluginUser::executeSkill(const skills::provider::dto::SkillExecutionInfo& info, const Ice::Current &current)
    {
        std::shared_lock l(skillsMutex);
        std::string skillName = info.skillName;
        ARMARX_CHECK_EXPRESSION(skillImplementations.count(skillName) > 0);

        auto& wrapper = skillImplementations.at(skillName);

        skills::SkillParameterization usedParameterization;
        usedParameterization.usedCallbackInterface = info.callbackInterface;
        usedParameterization.usedInputParams = aron::data::Dict::FromAronDictDTO(info.params);

        // We have to wait until the last task is finished
        if (wrapper.task.joinable())
        {
            wrapper.task.join();
        }

        // recreate thread and execute skill. A skill can only be executed once
        wrapper.task = std::thread{ [&] { wrapper.execute(usedParameterization);}};
        std::this_thread::sleep_for(std::chrono::milliseconds(50)); // somehow we need to wait, otherwise it chrashes


        if (info.waitUntilSkillFinished && wrapper.task.joinable())
        {
            // wait until task is finished. We hold the shared lock for the whole time.
            wrapper.task.join();
        }
    }

    void SkillProviderComponentPluginUser::abortSkill(const std::string& skillName, bool waitUntilSkillFinished, const Ice::Current &current)
    {
        std::shared_lock l(skillsMutex);
        ARMARX_CHECK_EXPRESSION(skillImplementations.count(skillName) > 0);

        auto& wrapper = skillImplementations.at(skillName);
        if (waitUntilSkillFinished && !wrapper.skill->stopRequested() && wrapper.task.joinable())
        {
            wrapper.skill->notifyStopped();
            wrapper.task.join();
        }
    }
}
